from bbot.modules.report.base import BaseReportModule


class aggregate(BaseReportModule):
    watched_events = []
    flags = ["passive", "safe"]
    meta = {
        "description": "Summarize statistics at the end of a scan",
        "created_date": "2022-07-25",
        "author": "@TheTechromancer",
    }

    async def report(self):
        self.log_table(*self.scan.stats._make_table(), table_name="scan-stats")



from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class c99(subdomain_enum_apikey):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query the C99 API for subdomains",
        "created_date": "2022-07-08",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "c99.nl API key"}

    base_url = "https://api.c99.nl"
    ping_url = f"{base_url}/randomnumber?key={{api_key}}&between=1,100&json"

    async def ping(self):
        url = f"{self.base_url}/randomnumber?key={{api_key}}&between=1,100&json"
        response = await self.api_request(url)
        assert response.json()["success"] is True, getattr(response, "text", "no response from server")

    async def request_url(self, query):
        url = f"{self.base_url}/subdomainfinder?key={{api_key}}&domain={self.helpers.quote(query)}&json"
        return await self.api_request(url)

    async def parse_results(self, r, query):
        results = set()
        j = r.json()
        if isinstance(j, dict):
            subdomains = j.get("subdomains", [])
            if subdomains:
                for s in subdomains:
                    subdomain = s.get("subdomain", "")
                    if subdomain:
                        results.add(subdomain)
        return results



from .base import ModuleTestBase


class TestPassiveTotal(ModuleTestBase):
    config_overrides = {"modules": {"passivetotal": {"api_key": "jon@bls.fakedomain:asdf"}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.passivetotal.org/v2/account/quota",
            match_headers={"Authorization": "Basic am9uQGJscy5mYWtlZG9tYWluOmFzZGY="},
            json={"user": {"counts": {"search_api": 10}, "limits": {"search_api": 20}}},
        )
        module_test.httpx_mock.add_response(
            url="https://api.passivetotal.org/v2/enrichment/subdomains?query=blacklanternsecurity.com",
            match_headers={"Authorization": "Basic am9uQGJscy5mYWtlZG9tYWluOmFzZGY="},
            json={"subdomains": ["asdf"]},
        )

    async def setup_after_prep(self, module_test):
        module_test.monkeypatch.setattr(module_test.scan.modules["passivetotal"], "abort_if", lambda e: False)

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class chaos(subdomain_enum_apikey):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query ProjectDiscovery's Chaos API for subdomains",
        "created_date": "2022-08-14",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "Chaos API key"}

    base_url = "https://dns.projectdiscovery.io/dns"
    ping_url = f"{base_url}/example.com"

    def prepare_api_request(self, url, kwargs):
        kwargs["headers"]["Authorization"] = self.api_key
        return url, kwargs

    async def request_url(self, query):
        _, domain = self.helpers.split_domain(query)
        url = f"{self.base_url}/{domain}/subdomains"
        return await self.api_request(url)

    async def parse_results(self, r, query):
        results = set()
        j = r.json()
        subdomains_set = set()
        if isinstance(j, dict):
            domain = j.get("domain", "")
            if domain:
                subdomains = j.get("subdomains", [])
                for s in subdomains:
                    s = s.lower().strip(".*")
                    subdomains_set.add(s)
                for s in subdomains_set:
                    full_subdomain = f"{s}.{domain}"
                    if full_subdomain and full_subdomain.endswith(f".{query}"):
                        results.add(full_subdomain)
        return results



from ..bbot_fixtures import *  # noqa: F401


@pytest.mark.asyncio
async def test_target_basic(bbot_scanner):
    from radixtarget import RadixTarget
    from ipaddress import ip_address, ip_network
    from bbot.scanner.target import BBOTTarget, ScanSeeds

    scan1 = bbot_scanner("api.publicapis.org", "8.8.8.8/30", "2001:4860:4860::8888/126")
    scan2 = bbot_scanner("8.8.8.8/29", "publicapis.org", "2001:4860:4860::8888/125")
    scan3 = bbot_scanner("8.8.8.8/29", "publicapis.org", "2001:4860:4860::8888/125")
    scan4 = bbot_scanner("8.8.8.8/29")
    scan5 = bbot_scanner()

    # test different types of inputs
    target = BBOTTarget("evilcorp.com", "1.2.3.4/8")
    assert "www.evilcorp.com" in target.seeds
    assert "www.evilcorp.com:80" in target.seeds
    assert "http://www.evilcorp.com:80" in target.seeds
    assert "1.2.3.4" in target.seeds
    assert "1.2.3.4/24" in target.seeds
    assert ip_address("1.2.3.4") in target.seeds
    assert ip_network("1.2.3.4/24", strict=False) in target.seeds
    event = scan1.make_event("https://www.evilcorp.com:80", dummy=True)
    assert event in target.seeds
    with pytest.raises(ValueError):
        ["asdf"] in target.seeds
    with pytest.raises(ValueError):
        target.seeds.get(["asdf"])

    assert not scan5.target.seeds
    assert len(scan1.target.seeds) == 9
    assert len(scan4.target.seeds) == 8
    assert "8.8.8.9" in scan1.target.seeds
    assert "8.8.8.12" not in scan1.target.seeds
    assert "8.8.8.8/31" in scan1.target.seeds
    assert "8.8.8.8/30" in scan1.target.seeds
    assert "8.8.8.8/29" not in scan1.target.seeds
    assert "2001:4860:4860::8889" in scan1.target.seeds
    assert "2001:4860:4860::888c" not in scan1.target.seeds
    assert "www.api.publicapis.org" in scan1.target.seeds
    assert "api.publicapis.org" in scan1.target.seeds
    assert "publicapis.org" not in scan1.target.seeds
    assert "bob@www.api.publicapis.org" in scan1.target.seeds
    assert "https://www.api.publicapis.org" in scan1.target.seeds
    assert "www.api.publicapis.org:80" in scan1.target.seeds
    assert scan1.make_event("https://[2001:4860:4860::8888]:80", dummy=True) in scan1.target.seeds
    assert scan1.make_event("[2001:4860:4860::8888]:80", "OPEN_TCP_PORT", dummy=True) in scan1.target.seeds
    assert scan1.make_event("[2001:4860:4860::888c]:80", "OPEN_TCP_PORT", dummy=True) not in scan1.target.seeds
    assert scan1.target.seeds in scan2.target.seeds
    assert scan2.target.seeds not in scan1.target.seeds
    assert scan3.target.seeds in scan2.target.seeds
    assert scan2.target.seeds == scan3.target.seeds
    assert scan4.target.seeds != scan1.target.seeds

    assert not scan5.target.whitelist
    assert len(scan1.target.whitelist) == 9
    assert len(scan4.target.whitelist) == 8
    assert "8.8.8.9" in scan1.target.whitelist
    assert "8.8.8.12" not in scan1.target.whitelist
    assert "8.8.8.8/31" in scan1.target.whitelist
    assert "8.8.8.8/30" in scan1.target.whitelist
    assert "8.8.8.8/29" not in scan1.target.whitelist
    assert "2001:4860:4860::8889" in scan1.target.whitelist
    assert "2001:4860:4860::888c" not in scan1.target.whitelist
    assert "www.api.publicapis.org" in scan1.target.whitelist
    assert "api.publicapis.org" in scan1.target.whitelist
    assert "publicapis.org" not in scan1.target.whitelist
    assert "bob@www.api.publicapis.org" in scan1.target.whitelist
    assert "https://www.api.publicapis.org" in scan1.target.whitelist
    assert "www.api.publicapis.org:80" in scan1.target.whitelist
    assert scan1.make_event("https://[2001:4860:4860::8888]:80", dummy=True) in scan1.target.whitelist
    assert scan1.make_event("[2001:4860:4860::8888]:80", "OPEN_TCP_PORT", dummy=True) in scan1.target.whitelist
    assert scan1.make_event("[2001:4860:4860::888c]:80", "OPEN_TCP_PORT", dummy=True) not in scan1.target.whitelist
    assert scan1.target.whitelist in scan2.target.whitelist
    assert scan2.target.whitelist not in scan1.target.whitelist
    assert scan3.target.whitelist in scan2.target.whitelist
    assert scan2.target.whitelist == scan3.target.whitelist
    assert scan4.target.whitelist != scan1.target.whitelist

    assert scan1.whitelisted("https://[2001:4860:4860::8888]:80")
    assert scan1.whitelisted("[2001:4860:4860::8888]:80")
    assert not scan1.whitelisted("[2001:4860:4860::888c]:80")
    assert scan1.whitelisted("www.api.publicapis.org")
    assert scan1.whitelisted("api.publicapis.org")
    assert not scan1.whitelisted("publicapis.org")

    assert scan1.target.seeds in scan2.target.seeds
    assert scan2.target.seeds not in scan1.target.seeds
    assert scan3.target.seeds in scan2.target.seeds
    assert scan2.target.seeds == scan3.target.seeds
    assert scan4.target.seeds != scan1.target.seeds

    assert str(scan1.target.seeds.get("8.8.8.9").host) == "8.8.8.8/30"
    assert str(scan1.target.whitelist.get("8.8.8.9").host) == "8.8.8.8/30"
    assert scan1.target.seeds.get("8.8.8.12") is None
    assert scan1.target.whitelist.get("8.8.8.12") is None
    assert str(scan1.target.seeds.get("2001:4860:4860::8889").host) == "2001:4860:4860::8888/126"
    assert str(scan1.target.whitelist.get("2001:4860:4860::8889").host) == "2001:4860:4860::8888/126"
    assert scan1.target.seeds.get("2001:4860:4860::888c") is None
    assert scan1.target.whitelist.get("2001:4860:4860::888c") is None
    assert str(scan1.target.seeds.get("www.api.publicapis.org").host) == "api.publicapis.org"
    assert str(scan1.target.whitelist.get("www.api.publicapis.org").host) == "api.publicapis.org"
    assert scan1.target.seeds.get("publicapis.org") is None
    assert scan1.target.whitelist.get("publicapis.org") is None

    target = RadixTarget("evilcorp.com")
    assert "com" not in target
    assert "evilcorp.com" in target
    assert "www.evilcorp.com" in target
    strict_target = RadixTarget("evilcorp.com", strict_dns_scope=True)
    assert "com" not in strict_target
    assert "evilcorp.com" in strict_target
    assert "www.evilcorp.com" not in strict_target

    target = RadixTarget()
    target.add("evilcorp.com")
    assert "com" not in target
    assert "evilcorp.com" in target
    assert "www.evilcorp.com" in target
    strict_target = RadixTarget(strict_dns_scope=True)
    strict_target.add("evilcorp.com")
    assert "com" not in strict_target
    assert "evilcorp.com" in strict_target
    assert "www.evilcorp.com" not in strict_target

    # test target hashing

    target1 = BBOTTarget()
    target1.whitelist.add("evilcorp.com")
    target1.whitelist.add("1.2.3.4/24")
    target1.whitelist.add("https://evilcorp.net:8080")
    target1.seeds.add("evilcorp.com")
    target1.seeds.add("1.2.3.4/24")
    target1.seeds.add("https://evilcorp.net:8080")

    target2 = BBOTTarget()
    target2.whitelist.add("bob@evilcorp.org")
    target2.whitelist.add("evilcorp.com")
    target2.whitelist.add("1.2.3.4/24")
    target2.whitelist.add("https://evilcorp.net:8080")
    target2.seeds.add("bob@evilcorp.org")
    target2.seeds.add("evilcorp.com")
    target2.seeds.add("1.2.3.4/24")
    target2.seeds.add("https://evilcorp.net:8080")

    # make sure it's a sha1 hash
    assert isinstance(target1.hash, bytes)
    assert len(target1.hash) == 20

    # hashes shouldn't match yet
    assert target1.hash != target2.hash
    assert target1.scope_hash != target2.scope_hash
    # add missing email
    target1.whitelist.add("bob@evilcorp.org")
    assert target1.hash != target2.hash
    assert target1.scope_hash == target2.scope_hash
    target1.seeds.add("bob@evilcorp.org")
    # now they should match
    assert target1.hash == target2.hash

    # test default whitelist
    bbottarget = BBOTTarget("http://1.2.3.4:8443", "bob@evilcorp.com")
    assert bbottarget.seeds.hosts == {ip_network("1.2.3.4"), "evilcorp.com"}
    assert bbottarget.whitelist.hosts == {ip_network("1.2.3.4"), "evilcorp.com"}
    assert {e.data for e in bbottarget.seeds.event_seeds} == {"http://1.2.3.4:8443/", "bob@evilcorp.com"}
    assert {e.data for e in bbottarget.whitelist.event_seeds} == {"1.2.3.4/32", "evilcorp.com"}

    bbottarget1 = BBOTTarget("evilcorp.com", "evilcorp.net", whitelist=["1.2.3.4/24"], blacklist=["1.2.3.4"])
    bbottarget2 = BBOTTarget("evilcorp.com", "evilcorp.net", whitelist=["1.2.3.0/24"], blacklist=["1.2.3.4"])
    bbottarget3 = BBOTTarget("evilcorp.com", whitelist=["1.2.3.4/24"], blacklist=["1.2.3.4"])
    bbottarget5 = BBOTTarget("evilcorp.com", "evilcorp.net", whitelist=["1.2.3.0/24"], blacklist=["1.2.3.4"])
    bbottarget6 = BBOTTarget(
        "evilcorp.com", "evilcorp.net", whitelist=["1.2.3.0/24"], blacklist=["1.2.3.4"], strict_scope=True
    )
    bbottarget8 = BBOTTarget("1.2.3.0/24", whitelist=["evilcorp.com", "evilcorp.net"], blacklist=["1.2.3.4"])
    bbottarget9 = BBOTTarget("evilcorp.com", "evilcorp.net", whitelist=["1.2.3.0/24"], blacklist=["1.2.3.4"])

    # make sure it's a sha1 hash
    assert isinstance(bbottarget1.hash, bytes)
    assert len(bbottarget1.hash) == 20

    assert bbottarget1 == bbottarget2
    assert bbottarget2 == bbottarget1
    # 1 and 3 have different seeds
    assert bbottarget1 != bbottarget3
    assert bbottarget3 != bbottarget1
    # until we make them the same
    bbottarget3.seeds.add("evilcorp.net")
    assert bbottarget1 == bbottarget3
    assert bbottarget3 == bbottarget1

    # adding different events (but with same host) to whitelist should not change hash (since only hosts matter)
    bbottarget1.whitelist.add("http://evilcorp.co.nz")
    bbottarget2.whitelist.add("evilcorp.co.nz")
    assert bbottarget1 == bbottarget2
    assert bbottarget2 == bbottarget1

    # but seeds should change hash
    bbottarget1.seeds.add("http://evilcorp.co.nz")
    bbottarget2.seeds.add("evilcorp.co.nz")
    assert bbottarget1 != bbottarget2
    assert bbottarget2 != bbottarget1

    # make sure strict_scope is considered in hash
    assert bbottarget5 != bbottarget6
    assert bbottarget6 != bbottarget5

    # make sure swapped target <--> whitelist result in different hash
    assert bbottarget8 != bbottarget9
    assert bbottarget9 != bbottarget8

    # make sure duplicate events don't change hash
    target1 = BBOTTarget("https://evilcorp.com")
    target2 = BBOTTarget("https://evilcorp.com")
    assert target1 == target2
    target1.seeds.add("https://evilcorp.com:443")
    assert target1 == target2

    # make sure hosts are collapsed in whitelist and blacklist
    bbottarget = BBOTTarget(
        "http://evilcorp.com:8080",
        whitelist=["evilcorp.net:443", "http://evilcorp.net:8080"],
        blacklist=["http://evilcorp.org:8080", "evilcorp.org:443"],
    )
    # base class is not iterable
    with pytest.raises(TypeError):
        assert list(bbottarget) == ["http://evilcorp.com:8080/"]
    assert {e.data for e in bbottarget.seeds} == {"http://evilcorp.com:8080/"}
    assert {e.data for e in bbottarget.whitelist} == {"evilcorp.net:443", "http://evilcorp.net:8080/"}
    assert {e.data for e in bbottarget.blacklist} == {"http://evilcorp.org:8080/", "evilcorp.org:443"}

    # test org stub as target
    for org_target in ("ORG:evilcorp", "ORG_STUB:evilcorp"):
        scan = bbot_scanner(org_target)
        events = [e async for e in scan.async_start()]
        assert len(events) == 3
        assert {e.type for e in events} == {"SCAN", "ORG_STUB"}

    # test username as target
    for user_target in ("USER:vancerefrigeration", "USERNAME:vancerefrigeration"):
        scan = bbot_scanner(user_target)
        events = [e async for e in scan.async_start()]
        assert len(events) == 3
        assert {e.type for e in events} == {"SCAN", "USERNAME"}

    # users + orgs + domains
    scan = bbot_scanner("USER:evilcorp", "ORG:evilcorp", "evilcorp.com")
    await scan.helpers.dns._mock_dns(
        {
            "evilcorp.com": {"A": ["1.2.3.4"]},
        },
    )
    events = [e async for e in scan.async_start()]
    assert len(events) == 5
    assert {e.type for e in events} == {"SCAN", "USERNAME", "ORG_STUB", "DNS_NAME"}

    # verify hash values
    bbottarget = BBOTTarget(
        "1.2.3.0/24",
        "http://www.evilcorp.net",
        "bob@fdsa.evilcorp.net",
        whitelist=["evilcorp.com", "bob@www.evilcorp.com", "evilcorp.net"],
        blacklist=["1.2.3.4", "4.3.2.1/24", "http://1.2.3.4", "bob@asdf.evilcorp.net"],
    )
    assert {e.data for e in bbottarget.seeds.event_seeds} == {
        "1.2.3.0/24",
        "http://www.evilcorp.net/",
        "bob@fdsa.evilcorp.net",
    }
    assert {e.data for e in bbottarget.whitelist.event_seeds} == {
        "evilcorp.com",
        "evilcorp.net",
        "bob@www.evilcorp.com",
    }
    assert {e.data for e in bbottarget.blacklist.event_seeds} == {
        "1.2.3.4",
        "4.3.2.0/24",
        "http://1.2.3.4/",
        "bob@asdf.evilcorp.net",
    }
    assert set(bbottarget.seeds.hosts) == {ip_network("1.2.3.0/24"), "www.evilcorp.net", "fdsa.evilcorp.net"}
    assert set(bbottarget.whitelist.hosts) == {"evilcorp.com", "evilcorp.net"}
    assert set(bbottarget.blacklist.hosts) == {ip_network("1.2.3.4/32"), ip_network("4.3.2.0/24"), "asdf.evilcorp.net"}
    assert bbottarget.hash == b"\xb3iU\xa8#\x8aq\x84/\xc5\xf2;\x11\x11\x0c&\xea\x07\xd4Q"
    assert bbottarget.scope_hash == b"f\xe1\x01c^3\xf5\xd24B\x87P\xa0Glq0p3J"
    assert bbottarget.seeds.hash == b"V\n\xf5\x1d\x1f=i\xbc\\\x15o\xc2p\xb2\x84\x97\xfeR\xde\xc1"
    assert bbottarget.whitelist.hash == b"\x8e\xd0\xa76\x8em4c\x0e\x1c\xfdA\x9d*sv}\xeb\xc4\xc4"
    assert bbottarget.blacklist.hash == b'\xf7\xaf\xa1\xda4"C:\x13\xf42\xc3,\xc3\xa9\x9f\x15\x15n\\'

    scan = bbot_scanner(
        "http://www.evilcorp.net",
        "1.2.3.0/24",
        "bob@fdsa.evilcorp.net",
        whitelist=["evilcorp.net", "evilcorp.com", "bob@www.evilcorp.com"],
        blacklist=["bob@asdf.evilcorp.net", "1.2.3.4", "4.3.2.1/24", "http://1.2.3.4"],
    )
    events = [e async for e in scan.async_start()]
    scan_events = [e for e in events if e.type == "SCAN"]
    assert len(scan_events) == 2
    target_dict = scan_events[0].data["target"]

    assert target_dict["seeds"] == ["1.2.3.0/24", "bob@fdsa.evilcorp.net", "http://www.evilcorp.net/"]
    assert target_dict["whitelist"] == ["bob@www.evilcorp.com", "evilcorp.com", "evilcorp.net"]
    assert target_dict["blacklist"] == ["1.2.3.4", "4.3.2.0/24", "bob@asdf.evilcorp.net", "http://1.2.3.4/"]
    assert target_dict["strict_scope"] is False
    assert target_dict["hash"] == "b36955a8238a71842fc5f23b11110c26ea07d451"
    assert target_dict["seed_hash"] == "560af51d1f3d69bc5c156fc270b28497fe52dec1"
    assert target_dict["whitelist_hash"] == "8ed0a7368e6d34630e1cfd419d2a73767debc4c4"
    assert target_dict["blacklist_hash"] == "f7afa1da3422433a13f432c32cc3a99f15156e5c"
    assert target_dict["scope_hash"] == "66e101635e33f5d234428750a0476c713070334a"

    # make sure child subnets/IPs don't get added to whitelist/blacklist
    target = RadixTarget("1.2.3.4/24", "1.2.3.4/28", acl_mode=True)
    assert set(target) == {ip_network("1.2.3.0/24")}
    target = RadixTarget("1.2.3.4/28", "1.2.3.4/24", acl_mode=True)
    assert set(target) == {ip_network("1.2.3.0/24")}
    target = RadixTarget("1.2.3.4/28", "1.2.3.4", acl_mode=True)
    assert set(target) == {ip_network("1.2.3.0/28")}
    target = RadixTarget("1.2.3.4", "1.2.3.4/28", acl_mode=True)
    assert set(target) == {ip_network("1.2.3.0/28")}

    # same but for domains
    target = RadixTarget("evilcorp.com", "www.evilcorp.com", acl_mode=True)
    assert set(target) == {"evilcorp.com"}
    target = RadixTarget("www.evilcorp.com", "evilcorp.com", acl_mode=True)
    assert set(target) == {"evilcorp.com"}

    # make sure strict_scope doesn't mess us up
    target = RadixTarget("evilcorp.co.uk", "www.evilcorp.co.uk", acl_mode=True, strict_dns_scope=True)
    assert set(target.hosts) == {"evilcorp.co.uk", "www.evilcorp.co.uk"}
    assert "evilcorp.co.uk" in target
    assert "www.evilcorp.co.uk" in target
    assert "api.evilcorp.co.uk" not in target
    assert "api.www.evilcorp.co.uk" not in target

    # test 'single' boolean argument
    target = ScanSeeds("http://evilcorp.com", "evilcorp.com:443")
    assert "www.evilcorp.com" in target
    assert "bob@evilcorp.com" in target
    event = target.get("www.evilcorp.com")
    assert event.host == "evilcorp.com"
    events = target.get("www.evilcorp.com", single=False)
    assert len(events) == 2
    assert {e.data for e in events} == {"http://evilcorp.com/", "evilcorp.com:443"}


@pytest.mark.asyncio
async def test_blacklist_regex(bbot_scanner, bbot_httpserver):
    from bbot.scanner.target import ScanBlacklist

    blacklist = ScanBlacklist("evilcorp.com")
    assert blacklist.inputs == {"evilcorp.com"}
    assert "www.evilcorp.com" in blacklist
    assert "http://www.evilcorp.com" in blacklist
    blacklist.add("RE:test")
    assert "REGEX:test" in blacklist.inputs
    assert set(blacklist.inputs) == {"evilcorp.com", "REGEX:test"}
    assert blacklist.blacklist_regexes
    assert next(iter(blacklist.blacklist_regexes)).pattern == "test"
    result1 = blacklist.get("test.com")
    assert result1 == "test.com"
    result2 = blacklist.get("www.evilcorp.com")
    assert result2 == "evilcorp.com"
    result2 = blacklist.get("www.evil.com")
    assert result2 is None
    with pytest.raises(KeyError):
        blacklist.get("www.evil.com", raise_error=True)
    assert "test.com" in blacklist
    assert "http://evilcorp.com/test.aspx" in blacklist
    assert "http://tes.com" not in blacklist

    blacklist = ScanBlacklist("evilcorp.com", r"RE:[0-9]{6}\.aspx$")
    assert "http://evilcorp.com" in blacklist
    assert "http://test.com/123456" not in blacklist
    assert "http://test.com/12345.aspx?a=asdf" not in blacklist
    assert "http://test.com/asdf/123456.aspx/asdf" not in blacklist
    assert "http://test.com/asdf/123456.aspx#asdf" in blacklist
    assert "http://test.com/asdf/123456.aspx" in blacklist

    bbot_httpserver.expect_request(uri="/").respond_with_data(
        """
        <a href='http://127.0.0.1:8888/asdfevil333asdf'/>
        <a href='http://127.0.0.1:8888/logout.aspx'/>
    """
    )
    bbot_httpserver.expect_request(uri="/asdfevilasdf").respond_with_data("")
    bbot_httpserver.expect_request(uri="/logout.aspx").respond_with_data("")

    # make sure URL is detected normally
    scan = bbot_scanner("http://127.0.0.1:8888/", presets=["spider"], config={"excavate": True}, debug=True)
    assert {r.pattern for r in scan.target.blacklist.blacklist_regexes} == {r"/.*(sign|log)[_-]?out"}
    events = [e async for e in scan.async_start()]
    urls = [e.data for e in events if e.type == "URL"]
    assert len(urls) == 2
    assert set(urls) == {"http://127.0.0.1:8888/", "http://127.0.0.1:8888/asdfevil333asdf"}

    # same scan again but with blacklist regex
    scan = bbot_scanner(
        "http://127.0.0.1:8888/",
        blacklist=[r"RE:evil[0-9]{3}"],
        presets=["spider"],
        config={"excavate": True},
        debug=True,
    )
    assert len(scan.target.blacklist) == 2
    assert scan.target.blacklist.blacklist_regexes
    assert {r.pattern for r in scan.target.blacklist.blacklist_regexes} == {
        r"evil[0-9]{3}",
        r"/.*(sign|log)[_-]?out",
    }
    events = [e async for e in scan.async_start()]
    urls = [e.data for e in events if e.type == "URL"]
    assert len(urls) == 1
    assert set(urls) == {"http://127.0.0.1:8888/"}



from .base import ModuleTestBase


class TestMySSL(ModuleTestBase):
    async def setup_after_prep(self, module_test):
        module_test.module.abort_if = lambda e: False
        module_test.httpx_mock.add_response(
            url="https://myssl.com/api/v1/discover_sub_domain?domain=blacklanternsecurity.com",
            json={
                "code": 0,
                "data": [
                    {
                        "ip": "1.2.3.4",
                        "port": "443",
                        "tips": [],
                        "level": 2,
                        "title": "",
                        "domain": "asdf.blacklanternsecurity.com",
                        "is_ats": True,
                        "is_pci": False,
                        "server": "",
                        "is_tlcp": False,
                        "duration": 46,
                        "icon_url": "",
                        "is_sslvpn": False,
                        "level_str": "A",
                        "ip_location": "美国",
                        "is_enable_gm": False,
                        "evaluate_date": "2022-03-13T02:38:08Z",
                        "demotion_reason": [],
                        "ignore_trust_level": "A",
                        "meet_gm_double_cert_statndard": False,
                    }
                ],
            },
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from .base import BaseLightfuzz
from bbot.errors import HttpCompareError

import statistics


class sqli(BaseLightfuzz):
    """
    Detects SQL injection vulnerabilities.

    Techniques:

    * Error-based Detection:
       - Injects single quotes and observes error responses
       - Tests quote escape sequence variations
       - Matches against known SQL error patterns

    * Time-based Blind Detection:
       - Uses vendor-specific time delay payloads
       - Confirms delays with statistical analysis
       - Requires multiple confirmations to eliminate false positives
    """

    friendly_name = "SQL Injection"

    expected_delay = 5
    # These are common error strings that strongly indicate SQL injection
    sqli_error_strings = [
        "Unterminated string literal",
        "Failed to parse string literal",
        "error in your SQL syntax",
        "syntax error at or near",
        "Unknown column",
        "unterminated quoted string",
        "Unclosed quotation mark",
        "Incorrect syntax near",
        "SQL command not properly ended",
        "string not properly terminated",
    ]

    def evaluate_delay(self, mean_baseline, measured_delay):
        """
        Evaluates if a measured delay falls within an expected range, indicating potential SQL injection.

        Parameters:
        - mean_baseline (float): The average baseline delay measured from non-injected requests.
        - measured_delay (float): The delay measured from a potentially injected request.

        Returns:
        - bool: True if the measured delay is within the expected range or exactly twice the expected delay, otherwise False.

        The function checks if the measured delay is within a margin of the expected delay or twice the expected delay,
        accounting for cases where the injected statement might be executed twice.
        """
        margin = 1.5
        if (
            mean_baseline + self.expected_delay - margin
            <= measured_delay
            <= mean_baseline + self.expected_delay + margin
        ):
            return True
        # check for exactly twice the delay, in case the statement gets placed in the query twice (a common occurrence)
        elif (
            mean_baseline + (self.expected_delay * 2) - margin
            <= measured_delay
            <= mean_baseline + (self.expected_delay * 2) + margin
        ):
            return True
        else:
            return False

    async def fuzz(self):
        cookies = self.event.data.get("assigned_cookies", {})
        probe_value = self.incoming_probe_value(populate_empty=True)
        http_compare = self.compare_baseline(
            self.event.data["type"], probe_value, cookies, additional_params_populate_empty=True
        )

        try:
            # send the with a single quote, and then another with two single quotes
            single_quote = await self.compare_probe(
                http_compare,
                self.event.data["type"],
                f"{probe_value}'",
                cookies,
                additional_params_populate_empty=True,
            )
            double_single_quote = await self.compare_probe(
                http_compare,
                self.event.data["type"],
                f"{probe_value}''",
                cookies,
                additional_params_populate_empty=True,
            )
            # if the single quote probe response is different from the baseline
            if single_quote[0] is False:
                # check for common SQL error strings in the response
                for sqli_error_string in self.sqli_error_strings:
                    if sqli_error_string.lower() in single_quote[3].text.lower():
                        self.results.append(
                            {
                                "type": "FINDING",
                                "description": f"Possible SQL Injection. {self.metadata()} Detection Method: [SQL Error Detection] Detected String: [{sqli_error_string}]",
                            }
                        )
                        break
            # if both probes were successful (and had a response)
            if single_quote[3] and double_single_quote[3]:
                # Ensure none of the status codes are "429"
                if (
                    single_quote[3].status_code != 429
                    and double_single_quote[3].status_code != 429
                    and http_compare.baseline.status_code != 429
                    and http_compare.baseline.status_code != 403  # Ensure the baseline status code is not 403
                ):  # prevent false positives from rate limiting
                    # if the code changed in the single quote probe, and the code is NOT the same between that and the double single quote probe, SQL injection is indicated
                    if "code" in single_quote[1] and (
                        single_quote[3].status_code != double_single_quote[3].status_code
                    ):
                        self.results.append(
                            {
                                "type": "FINDING",
                                "description": f"Possible SQL Injection. {self.metadata()} Detection Method: [Single Quote/Two Single Quote, Code Change ({http_compare.baseline.status_code}->{single_quote[3].status_code}->{double_single_quote[3].status_code})]",
                            }
                        )
            else:
                self.debug("Failed to get responses for both single_quote and double_single_quote")
        except HttpCompareError as e:
            self.verbose(f"Encountered HttpCompareError Sending Compare Probe: {e}")

        # These are common SQL injection payloads for inducing an intentional delay across several different SQL database types
        standard_probe_strings = [
            f"'||pg_sleep({str(self.expected_delay)})--",  # postgres
            f"1' AND (SLEEP({str(self.expected_delay)})) AND '",  # mysql
            f"' AND (SELECT FROM DBMS_LOCK.SLEEP({str(self.expected_delay)})) AND '1'='1"  # oracle (not tested)
            f"; WAITFOR DELAY '00:00:{str(self.expected_delay)}'--",  # mssql (not tested)
        ]

        baseline_1 = await self.standard_probe(
            self.event.data["type"], cookies, probe_value, additional_params_populate_empty=True
        )
        baseline_2 = await self.standard_probe(
            self.event.data["type"], cookies, probe_value, additional_params_populate_empty=True
        )

        # get a baseline from two different probes. We will average them to establish a mean baseline
        if baseline_1 and baseline_2:
            baseline_1_delay = baseline_1.elapsed.total_seconds()
            baseline_2_delay = baseline_2.elapsed.total_seconds()
            mean_baseline = statistics.mean([baseline_1_delay, baseline_2_delay])

            for p in standard_probe_strings:
                confirmations = 0
                for i in range(0, 3):
                    # send the probe 3 times, and check if the delay is within the detection threshold
                    r = await self.standard_probe(
                        self.event.data["type"],
                        cookies,
                        f"{probe_value}{p}",
                        additional_params_populate_empty=True,
                        timeout=20,
                    )
                    if not r:
                        self.debug("delay measure request failed")
                        break

                    d = r.elapsed.total_seconds()
                    self.debug(f"measured delay: {str(d)}")
                    if self.evaluate_delay(
                        mean_baseline, d
                    ):  # decide if the delay is within the detection threshold and constitutes a successful sleep execution
                        confirmations += 1
                        self.debug(
                            f"{self.event.data['url']}:{self.event.data['name']}:{self.event.data['type']} Increasing confirmations, now: {str(confirmations)} "
                        )
                    else:
                        break

                if confirmations == 3:
                    self.results.append(
                        {
                            "type": "FINDING",
                            "description": f"Possible Blind SQL Injection. {self.metadata()} Detection Method: [Delay Probe ({p})]",
                        }
                    )

        else:
            self.debug("Could not get baseline for time-delay tests")



from .base import ModuleTestBase


class TestPGP(ModuleTestBase):
    web_body = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd" >
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<title>Search results for 'blacklanternsecurity.com'</title>
<meta http-equiv="Content-Type" content="text/html;charset=utf-8" />
<link href='/assets/css/pks.min.css' rel='stylesheet' type='text/css'>
<style type="text/css">

 .uid { color: green; text-decoration: underline; }
 .warn { color: red; font-weight: bold; }

</style></head><body><h1>Search results for 'blacklanternsecurity.com'</h1><pre>Type bits/keyID            cr. time   exp time   key expir
</pre>


<hr /><pre><strong>pub</strong> <a href="/pks/lookup?op=get&search=0xd4e98af823deadbeef">eddsa263/0xd4e98af823deadbeef</a> 2022-09-14T15:11:31Z

<strong>uid</strong> <span class="uid">Asdf &lt;asdf@blacklanternsecurity.com&gt;</span>
sig  sig  <a href="/pks/lookup?op=get&search=0xd4e98af823deadbeef">0xd4e98af823deadbeef</a> 2022-09-14T15:11:31Z 2024-09-14T17:00:00Z ____________________ <a href="/pks/lookup?op=vindex&search=0xd4e98af823deadbeef">[selfsig]</a>

</pre>
</body></html>"""

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://keyserver.ubuntu.com/pks/lookup?fingerprint=on&op=vindex&search=blacklanternsecurity.com",
            text=self.web_body,
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf@blacklanternsecurity.com" for e in events), "Failed to detect email"



# dnstlsrpt.py
#
# Checks for and parses common TLS-RPT TXT records, e.g. _smtp._tls.target.domain
#
# TLS-RPT policies may contain email addresses or URL's for reporting destinations, typically the email addresses are software processed inboxes, but they may also be to individual humans or team inboxes.
#
# The domain portion of any email address or URL is also passively checked and added as appropriate, for additional inspection by other modules.
#
# Example records,
# _smtp._tls.example.com TXT "v=TLSRPTv1;rua=https://tlsrpt.azurewebsites.net/report"
# _smtp._tls.example.net TXT "v=TLSRPTv1; rua=mailto:sts-reports@example.net;"
#
# TODO: extract %{UNIQUE_ID}% from hosted services as ORG_STUB ?
#   e.g. %{UNIQUE_ID}%@tlsrpt.hosted.service.provider is usually a tenant specific ID.
#   e.g. tlsrpt@%{UNIQUE_ID}%.hosted.service.provider is usually a tenant specific ID.

from bbot.modules.base import BaseModule
from bbot.core.helpers.dns.helpers import service_record

import re

from bbot.core.helpers.regexes import email_regex, url_regexes

_tlsrpt_regex = r"^v=(?P<v>TLSRPTv[0-9]+); *(?P<kvps>.*)$"
tlsrpt_regex = re.compile(_tlsrpt_regex, re.I)

_tlsrpt_kvp_regex = r"(?P<k>\w+)=(?P<v>[^;]+);*"
tlsrpt_kvp_regex = re.compile(_tlsrpt_kvp_regex)

_csul = r"(?P<uri>[^, ]+)"
csul = re.compile(_csul)


class dnstlsrpt(BaseModule):
    watched_events = ["DNS_NAME"]
    produced_events = ["EMAIL_ADDRESS", "URL_UNVERIFIED", "RAW_DNS_RECORD"]
    flags = ["subdomain-enum", "cloud-enum", "email-enum", "passive", "safe"]
    meta = {
        "description": "Check for TLS-RPT records",
        "author": "@colin-stubbs",
        "created_date": "2024-07-26",
    }
    options = {
        "emit_emails": True,
        "emit_raw_dns_records": False,
        "emit_urls": True,
        "emit_vulnerabilities": True,
    }
    options_desc = {
        "emit_emails": "Emit EMAIL_ADDRESS events",
        "emit_raw_dns_records": "Emit RAW_DNS_RECORD events",
        "emit_urls": "Emit URL_UNVERIFIED events",
        "emit_vulnerabilities": "Emit VULNERABILITY events",
    }

    async def setup(self):
        self.emit_emails = self.config.get("emit_emails", True)
        self.emit_raw_dns_records = self.config.get("emit_raw_dns_records", False)
        self.emit_urls = self.config.get("emit_urls", True)
        self.emit_vulnerabilities = self.config.get("emit_vulnerabilities", True)
        return await super().setup()

    def _incoming_dedup_hash(self, event):
        # dedupe by parent
        parent_domain = self.helpers.parent_domain(event.data)
        return hash(parent_domain), "already processed parent domain"

    async def filter_event(self, event):
        if "_wildcard" in str(event.host).split("."):
            return False, "event is wildcard"

        # there's no value in inspecting service records
        if service_record(event.host) is True:
            return False, "service record detected"

        return True

    async def handle_event(self, event):
        rdtype = "TXT"
        tags = ["tlsrpt-record"]
        hostname = f"_smtp._tls.{event.host}"

        r = await self.helpers.resolve_raw(hostname, type=rdtype)

        if r:
            raw_results, errors = r
            for answer in raw_results:
                if self.emit_raw_dns_records:
                    await self.emit_event(
                        {"host": hostname, "type": rdtype, "answer": answer.to_text()},
                        "RAW_DNS_RECORD",
                        parent=event,
                        tags=tags.append(f"{rdtype.lower()}-record"),
                        context=f"{rdtype} lookup on {hostname} produced {{event.type}}",
                    )

                # we need to fix TXT data that may have been split across two different rdata's
                # e.g. we will get a single string, but within that string we may have two parts such as:
                # answer = '"part 1 that was really long" "part 2 that did not fit in part 1"'
                # NOTE: the leading and trailing double quotes are essential as part of a raw DNS TXT record, or another record type that contains a free form text string as a component.
                s = answer.to_text().strip('"').replace('" "', "")

                # validate TLSRPT record, tag appropriately
                tlsrpt_match = tlsrpt_regex.search(s)

                if (
                    tlsrpt_match
                    and tlsrpt_match.group("v")
                    and tlsrpt_match.group("kvps")
                    and tlsrpt_match.group("kvps") != ""
                ):
                    for kvp_match in tlsrpt_kvp_regex.finditer(tlsrpt_match.group("kvps")):
                        key = kvp_match.group("k").lower()

                        if key == "rua":
                            for csul_match in csul.finditer(kvp_match.group("v")):
                                if csul_match.group("uri"):
                                    for match in email_regex.finditer(csul_match.group("uri")):
                                        start, end = match.span()
                                        email = csul_match.group("uri")[start:end]

                                        if self.emit_emails:
                                            await self.emit_event(
                                                email,
                                                "EMAIL_ADDRESS",
                                                tags=tags.append(f"tlsrpt-record-{key}"),
                                                parent=event,
                                            )

                                    for url_regex in url_regexes:
                                        for match in url_regex.finditer(csul_match.group("uri")):
                                            start, end = match.span()
                                            url = csul_match.group("uri")[start:end]

                                            if self.emit_urls:
                                                await self.emit_event(
                                                    url,
                                                    "URL_UNVERIFIED",
                                                    tags=tags.append(f"tlsrpt-record-{key}"),
                                                    parent=event,
                                                )


# EOF



#!/usr/bin/env python3

import io
import sys
import logging
import multiprocessing
from bbot.errors import *
from bbot import __version__
from bbot.logger import log_to_stderr
from bbot.core.helpers.misc import chain_lists


if multiprocessing.current_process().name == "MainProcess":
    silent = "-s" in sys.argv or "--silent" in sys.argv

    if not silent:
        ascii_art = rf""" [1;38;5;208m ______ [0m _____   ____ _______
 [1;38;5;208m|  ___ \[0m|  __ \ / __ \__   __|
 [1;38;5;208m| |___) [0m| |__) | |  | | | |
 [1;38;5;208m|  ___ <[0m|  __ <| |  | | | |
 [1;38;5;208m| |___) [0m| |__) | |__| | | |
 [1;38;5;208m|______/[0m|_____/ \____/  |_|
 [1;38;5;208mBIGHUGE[0m BLS OSINT TOOL {__version__}

www.blacklanternsecurity.com/bbot
"""
        print(ascii_art, file=sys.stderr)

scan_name = ""


async def _main():
    import asyncio
    import traceback
    from contextlib import suppress

    # fix tee buffering
    sys.stdout.reconfigure(line_buffering=True)

    log = logging.getLogger("bbot.cli")

    from bbot.scanner import Scanner
    from bbot.scanner.preset import Preset

    global scan_name

    try:
        # start by creating a default scan preset
        preset = Preset(_log=True, name="bbot_cli_main")
        # parse command line arguments and merge into preset
        try:
            preset.parse_args()
        except BBOTArgumentError as e:
            log_to_stderr(str(e), level="WARNING")
            log.trace(traceback.format_exc())
            return
        # ensure arguments (-c config options etc.) are valid
        options = preset.args.parsed

        # print help if no arguments
        if len(sys.argv) == 1:
            print(preset.args.parser.format_help())
            sys.exit(1)
            return

        # --version
        if options.version:
            print(__version__)
            sys.exit(0)
            return

        # --list-presets
        if options.list_presets:
            print("")
            print("### PRESETS ###")
            print("")
            for row in preset.presets_table().splitlines():
                print(row)
            return

        # if we're listing modules or their options
        if options.list_modules or options.list_output_modules or options.list_module_options or options.module_help:
            # if no modules or flags are specified, enable everything
            if not (options.modules or options.output_modules or options.flags):
                for module, preloaded in preset.module_loader.preloaded().items():
                    module_type = preloaded.get("type", "scan")
                    preset.add_module(module, module_type=module_type)

            if options.modules or options.output_modules or options.flags:
                preset._default_output_modules = options.output_modules
                preset._default_internal_modules = []

            preset.bake()

            # --list-modules
            if options.list_modules:
                print("")
                print("### MODULES ###")
                print("")
                modules = sorted(set(preset.scan_modules + preset.internal_modules))
                for row in preset.module_loader.modules_table(modules).splitlines():
                    print(row)
                return

            # --list-output-modules
            if options.list_output_modules:
                print("")
                print("### OUTPUT MODULES ###")
                print("")
                for row in preset.module_loader.modules_table(preset.output_modules).splitlines():
                    print(row)
                return

            # --list-module-options
            if options.list_module_options:
                print("")
                print("### MODULE OPTIONS ###")
                print("")
                for row in preset.module_loader.modules_options_table(preset.modules).splitlines():
                    print(row)
                return

            # --module-help
            if options.module_help:
                module_name = options.module_help
                all_modules = list(preset.module_loader.preloaded())
                if module_name not in all_modules:
                    log.hugewarning(f'Module "{module_name}" not found')
                    return

                # Load the module class
                loaded_modules = preset.module_loader.load_modules([module_name])
                module_name, module_class = next(iter(loaded_modules.items()))
                print(module_class.help_text())
                return

        # --list-flags
        if options.list_flags:
            flags = preset.flags if preset.flags else None
            print("")
            print("### FLAGS ###")
            print("")
            for row in preset.module_loader.flags_table(flags=flags).splitlines():
                print(row)
            return

        try:
            scan = Scanner(preset=preset)
        except (PresetAbortError, ValidationError) as e:
            log.warning(str(e))
            return

        deadly_modules = [
            m for m in scan.preset.scan_modules if "deadly" in preset.preloaded_module(m).get("flags", [])
        ]
        if deadly_modules and not options.allow_deadly:
            log.hugewarning(f"You enabled the following deadly modules: {','.join(deadly_modules)}")
            log.hugewarning("Deadly modules are highly intrusive")
            log.hugewarning("Please specify --allow-deadly to continue")
            return False

        # --current-preset
        if options.current_preset:
            print(scan.preset.to_yaml())
            sys.exit(0)
            return

        # --current-preset-full
        if options.current_preset_full:
            print(scan.preset.to_yaml(full_config=True))
            sys.exit(0)
            return

        # --install-all-deps
        if options.install_all_deps:
            all_modules = list(preset.module_loader.preloaded())
            scan.helpers.depsinstaller.force_deps = True
            succeeded, failed = await scan.helpers.depsinstaller.install(*all_modules)
            if failed:
                log.hugewarning(f"Failed to install dependencies for the following modules: {', '.join(failed)}")
                return False
            log.hugesuccess(f"Successfully installed dependencies for the following modules: {', '.join(succeeded)}")
            return True

        scan_name = str(scan.name)

        log.verbose("")
        log.verbose("### MODULES ENABLED ###")
        log.verbose("")
        for row in scan.preset.module_loader.modules_table(scan.preset.modules).splitlines():
            log.verbose(row)

        scan.helpers.word_cloud.load()
        await scan._prep()

        if not options.dry_run:
            log.trace(f"Command: {' '.join(sys.argv)}")

            if sys.stdin.isatty():
                # warn if any targets belong directly to a cloud provider
                if not scan.preset.strict_scope:
                    for event in scan.target.seeds.event_seeds:
                        if event.type == "DNS_NAME":
                            cloudcheck_result = scan.helpers.cloudcheck(event.host)
                            if cloudcheck_result:
                                scan.hugewarning(
                                    f'YOUR TARGET CONTAINS A CLOUD DOMAIN: "{event.host}". You\'re in for a wild ride!'
                                )

                if not options.yes:
                    log.hugesuccess(f"Scan ready. Press enter to execute {scan.name}")
                    input()

                import os
                import re
                import fcntl
                from bbot.core.helpers.misc import smart_decode

                def handle_keyboard_input(keyboard_input):
                    kill_regex = re.compile(r"kill (?P<modules>[a-z0-9_ ,]+)")
                    if keyboard_input:
                        log.verbose(f'Got keyboard input: "{keyboard_input}"')
                        kill_match = kill_regex.match(keyboard_input)
                        if kill_match:
                            modules = kill_match.group("modules")
                            if modules:
                                modules = chain_lists(modules)
                                for module in modules:
                                    if module in scan.modules:
                                        log.hugewarning(f'Killing module: "{module}"')
                                        scan.kill_module(module, message="killed by user")
                                    else:
                                        log.warning(f'Invalid module: "{module}"')
                    else:
                        scan.preset.core.logger.toggle_log_level(logger=log)
                        scan.modules_status(_log=True)

                reader = asyncio.StreamReader()
                protocol = asyncio.StreamReaderProtocol(reader)
                await asyncio.get_running_loop().connect_read_pipe(lambda: protocol, sys.stdin)

                # set stdout and stderr to blocking mode
                # this is needed to prevent BlockingIOErrors in logging etc.
                fds = []
                for stream in [sys.stdout, sys.stderr]:
                    try:
                        fds.append(stream.fileno())
                    except io.UnsupportedOperation:
                        log.debug(f"Can't get fileno for {stream}")
                for fd in fds:
                    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
                    fcntl.fcntl(fd, fcntl.F_SETFL, flags & ~os.O_NONBLOCK)

                async def akeyboard_listen():
                    try:
                        allowed_errors = 10
                        while 1:
                            keyboard_input = None
                            try:
                                keyboard_input = smart_decode((await reader.readline()).strip())
                                allowed_errors = 10
                            except Exception as e:
                                log_to_stderr(f"Error in keyboard listen loop: {e}", level="TRACE")
                                log_to_stderr(traceback.format_exc(), level="TRACE")
                                allowed_errors -= 1
                            if keyboard_input is not None:
                                handle_keyboard_input(keyboard_input)
                            if allowed_errors <= 0:
                                break
                    except Exception as e:
                        log_to_stderr(f"Error in keyboard listen task: {e}", level="ERROR")
                        log_to_stderr(traceback.format_exc(), level="TRACE")

                keyboard_listen_task = asyncio.create_task(akeyboard_listen())  # noqa F841

            await scan.async_start_without_generator()

        return True

    except BBOTError as e:
        log.error(str(e))
        log.trace(traceback.format_exc())

    finally:
        # save word cloud
        with suppress(BaseException):
            scan.helpers.word_cloud.save()
        # remove output directory if empty
        with suppress(BaseException):
            scan.home.rmdir()


def main():
    import asyncio
    import traceback
    from bbot.core import CORE

    global scan_name
    try:
        asyncio.run(_main())
    except asyncio.CancelledError:
        if CORE.logger.log_level <= logging.DEBUG:
            log_to_stderr(traceback.format_exc(), level="DEBUG")
    except KeyboardInterrupt:
        msg = "Interrupted"
        if scan_name:
            msg = f"You killed {scan_name}"
        log_to_stderr(msg, level="WARNING")
        if CORE.logger.log_level <= logging.DEBUG:
            log_to_stderr(traceback.format_exc(), level="DEBUG")
        exit(1)


if __name__ == "__main__":
    main()



import time
import pytest
import string
import random


@pytest.mark.asyncio
async def test_bloom_filter():
    def generate_random_strings(n, length=10):
        """Generate a list of n random strings."""
        return ["".join(random.choices(string.ascii_letters + string.digits, k=length)) for _ in range(n)]

    from bbot.scanner import Scanner

    scan = Scanner()

    n_items_to_add = 100000
    n_items_to_test = 100000
    bloom_filter_size = 8000000

    # Initialize the simple bloom filter and the set
    bloom_filter = scan.helpers.bloom_filter(size=bloom_filter_size)

    test_set = set()

    # Generate random strings to add
    print(f"Generating {n_items_to_add:,} items to add")
    items_to_add = set(generate_random_strings(n_items_to_add))

    # Generate random strings to test
    print(f"Generating {n_items_to_test:,} items to test")
    items_to_test = generate_random_strings(n_items_to_test)

    print("Adding items")
    start = time.time()
    for item in items_to_add:
        bloom_filter.add(item)
        test_set.add(hash(item))
    end = time.time()
    elapsed = end - start
    print(f"elapsed: {elapsed:.2f} ({int(n_items_to_test / elapsed)}/s)")
    # this shouldn't take longer than 5 seconds
    assert elapsed < 5

    # make sure we have 100% accuracy
    start = time.time()
    for item in items_to_add:
        assert item in bloom_filter
    end = time.time()
    elapsed = end - start
    print(f"elapsed: {elapsed:.2f} ({int(n_items_to_test / elapsed)}/s)")
    # this shouldn't take longer than 5 seconds
    assert elapsed < 5

    print("Measuring false positives")
    # Check for false positives
    false_positives = 0
    for item in items_to_test:
        if bloom_filter.check(item) and hash(item) not in test_set:
            false_positives += 1
    false_positive_percent = false_positives / len(items_to_test) * 100

    print(f"False positive rate: {false_positive_percent:.2f}% ({false_positives}/{len(items_to_test)})")

    # ensure false positives are less than .02 percent
    assert false_positive_percent < 0.02

    bloom_filter.close()

    await scan._cleanup()



from bbot.modules.base import BaseModule


class postman(BaseModule):
    """
    A template module for use of the GitHub API
    Inherited by several other github modules.
    """

    base_url = "https://www.postman.com/_api"
    api_url = "https://api.getpostman.com"
    html_url = "https://www.postman.com"
    ping_url = f"{api_url}/me"

    headers = {
        "Content-Type": "application/json",
        "X-App-Version": "11.27.4-250109-2338",
        "X-Entity-Team-Id": "0",
        "Origin": "https://www.postman.com",
        "Referer": "https://www.postman.com/search?q=&scope=public&type=all",
    }
    auth_required = True

    async def setup(self):
        await super().setup()
        self.headers = {}
        api_keys = set()
        modules_config = self.scan.config.get("modules", {})
        postman_modules = [m for m in modules_config if str(m).startswith("postman")]
        for module_name in postman_modules:
            module_config = modules_config.get(module_name, {})
            api_key = module_config.get("api_key", "")
            if isinstance(api_key, str):
                api_key = [api_key]
            for key in api_key:
                key = key.strip()
                if key:
                    api_keys.add(key)
        if not api_keys:
            if self.auth_required:
                return None, "No API key set"
        self.api_key = api_keys
        if self.api_key:
            try:
                await self.ping()
                self.hugesuccess("API is ready")
                return True
            except Exception as e:
                self.trace()
                return None, f"Error with API ({str(e).strip()})"
        return True

    def prepare_api_request(self, url, kwargs):
        if self.api_key:
            kwargs["headers"]["X-Api-Key"] = self.api_key
        return url, kwargs

    async def get_workspace_id(self, repo_url):
        workspace_id = ""
        profile = repo_url.split("/")[-2]
        name = repo_url.split("/")[-1]
        url = f"{self.base_url}/ws/proxy"
        json = {
            "service": "workspaces",
            "method": "GET",
            "path": f"/workspaces?handle={profile}&slug={name}",
        }
        r = await self.helpers.request(url, method="POST", json=json, headers=self.headers)
        if r is None:
            return workspace_id
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return workspace_id
        data = json.get("data", [])
        if len(data) == 1:
            workspace_id = data[0]["id"]
        return workspace_id

    async def request_workspace(self, id):
        data = {"workspace": {}, "environments": [], "collections": []}
        workspace = await self.get_workspace(id)
        if workspace:
            # Main Workspace
            name = workspace["name"]
            data["workspace"] = workspace

            # Workspace global variables
            self.verbose(f"Searching globals for workspace {name}")
            globals = await self.get_globals(id)
            data["environments"].append(globals)

            # Workspace Environments
            workspace_environments = workspace.get("environments", [])
            if workspace_environments:
                self.verbose(f"Searching environments for workspace {name}")
                for _ in workspace_environments:
                    environment_id = _["uid"]
                    environment = await self.get_environment(environment_id)
                    data["environments"].append(environment)

            # Workspace Collections
            workspace_collections = workspace.get("collections", [])
            if workspace_collections:
                self.verbose(f"Searching collections for workspace {name}")
                for _ in workspace_collections:
                    collection_id = _["uid"]
                    collection = await self.get_collection(collection_id)
                    data["collections"].append(collection)
        return data

    async def get_workspace(self, workspace_id):
        workspace = {}
        workspace_url = f"{self.api_url}/workspaces/{workspace_id}"
        r = await self.api_request(workspace_url)
        if r is None:
            return workspace
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return workspace
        workspace = json.get("workspace", {})
        return workspace

    async def get_globals(self, workspace_id):
        globals = {}
        globals_url = f"{self.base_url}/workspace/{workspace_id}/globals"
        r = await self.helpers.request(globals_url, headers=self.headers)
        if r is None:
            return globals
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return globals
        globals = json.get("data", {})
        return globals

    async def get_environment(self, environment_id):
        environment = {}
        environment_url = f"{self.api_url}/environments/{environment_id}"
        r = await self.api_request(environment_url)
        if r is None:
            return environment
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return environment
        environment = json.get("environment", {})
        return environment

    async def get_collection(self, collection_id):
        collection = {}
        collection_url = f"{self.api_url}/collections/{collection_id}"
        r = await self.api_request(collection_url)
        if r is None:
            return collection
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return collection
        collection = json.get("collection", {})
        return collection

    async def validate_workspace(self, workspace, environments, collections):
        name = workspace.get("name", "")
        full_wks = str([workspace, environments, collections])
        in_scope_hosts = await self.scan.extract_in_scope_hostnames(full_wks)
        if in_scope_hosts:
            self.verbose(
                f'Found in-scope hostname(s): "{in_scope_hosts}" in workspace {name}, it appears to be in-scope'
            )
            return True
        return False



import re

from .base import ModuleTestBase


class TestIIS_Shortnames(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "iis_shortnames"]
    config_overrides = {"modules": {"iis_shortnames": {"detect_only": False}}}

    async def setup_after_prep(self, module_test):
        module_test.httpserver.no_handler_status_code = 404

        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": "alive", "status": 200}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/*~1*/a.aspx"}
        respond_args = {"response_data": "", "status": 400}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": re.compile(r"\/B\*~1\*.*$")}
        respond_args = {"response_data": "", "status": 400}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": re.compile(r"\/BL\*~1\*.*$")}
        respond_args = {"response_data": "", "status": 400}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": re.compile(r"\/BLS\*~1\*.*$")}
        respond_args = {"response_data": "", "status": 400}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": re.compile(r"\/BLSH\*~1\*.*$")}
        respond_args = {"response_data": "", "status": 400}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": re.compile(r"\/BLSHA\*~1\*.*$")}
        respond_args = {"response_data": "", "status": 400}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": re.compile(r"\/BLSHAX\*~1\*.*$")}
        respond_args = {"response_data": "", "status": 400}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        for char in "BLSHAX":
            expect_args = {"method": "GET", "uri": re.compile(rf"\/\*{char}\*~1\*.*$")}
            respond_args = {"response_data": "", "status": 400}
            module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        vulnerabilityEmitted = False
        url_hintEmitted = False
        for e in events:
            if e.type == "VULNERABILITY" and "iis-magic-url" not in e.tags:
                vulnerabilityEmitted = True
            if e.type == "URL_HINT" and e.data == "http://127.0.0.1:8888/BLSHAX~1":
                url_hintEmitted = True

        assert vulnerabilityEmitted
        assert url_hintEmitted
